#include <linux/config.h>

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#endif

#include <linux/ioport.h>	/* request_region */
#include <linux/malloc.h>
#include <linux/pci.h>		/* PCI routine */
#include <linux/errno.h>
#include <asm/io.h>			/* inl, outl, ... */

#include <rtl_conf.h>
#include <rtl_core.h>
#include <rtl_sync.h>
#include <rtl_posixio.h>
#define	RTL_K2915_MAJOR	30

#include "cc_2915.h"


#ifdef DEBUG
  	#define DEBUG2915 1
#else
  	/* #define DEBUG2915 1 */
#endif

#define K2915_CNT	1

struct cc_struct {
  	int base1;						/* IOR Register */
  	int base2;						/* PBOR Register*/
  	unsigned int irq;				/* IRQ number 	*/
  	int status;						/* Crate status */
  	short cur_branch;				/* CAMAC branch	*/
  	short cur_crate;				/* CAMAC crate  */
  	struct wait_queue *cc_wait_q;
	};

struct cc_struct cc_table[K2915_CNT] =
 {
  	{0, 0, 0, CC_K_READY, 0, 0, NULL,}
 };

struct  camac {
  	/* Parallel Bus Operational Register */
  	unsigned int csr;	/* Control/Statues Register 	*/
  	unsigned int cnaf;	/* CNAF Register 				*/
  	unsigned int tcr;	/* Transfer Count Register		*/
  	unsigned int srr;	/* Service Request Register		*/
  	/* PCI Interface Operational Register */
  	unsigned int fifo;	/* Data FIFO					*/
  	unsigned int mwa;	/* Master Write Address			*/
  	unsigned int mwtc;	/* Master Write Transfer Count	*/
  	unsigned int mra;	/* Master Read Address			*/
  	unsigned int mrtc;	/* Master Read Transfer Count	*/
  	unsigned int ics;	/* Interrupt Control/Status		*/
  	unsigned int bmcs;  /* Bus Master Control/Status	*/
	};

struct camac reg[K2915_CNT] =
 {
   	{CSR, CNAF, TCR, SRR, FIFO, MWA, MWTC, MRA, MRTC, ICS, BMCS,},
 };

static int cc_probe(struct cc_struct *cc_pci_hdr, struct camac *cc_reg);

static int cc_open (struct rtl_file *file)
 {
	rtl_printf("CAMAC:open.\n");
  	return 0;
 }

static int cc_release (struct rtl_file *file)
 {
	rtl_printf("CAMAC:close.\n");
  	return 0;
 }

static struct rtl_file_operations cc_fops = {
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	cc_open,
	cc_release
	};

#ifdef MODULE
int init_module(void)
{
  	struct cc_struct 	*cc = &cc_table[0];

  	if (! cc_probe(&cc_table[0], &reg[0]) ) {
   		cc = &(cc_table[0]);

   		if((check_region(cc->base1,IOR_SIZE ) != 0) ||
		   (check_region(cc->base2,PBOR_SIZE) != 0) )
    		{
      		rtl_printf("CAMAC: No space for K2915 registers\n");
      		return -EIO;
    		}
   		request_region(cc->base1,IOR_SIZE, "CAMAC K2915(IOR)" );
   		request_region(cc->base2,PBOR_SIZE,"CAMAC K2915(PBOR)");
    
   		rtl_printf("K2915 CAMAC Device Driver was Installed:\n");
   		/*
		rtl_printf("   IRQ : 0x%8.8x\n",cc->irq);
   		rtl_printf("   IOR : 0x%8.8x\n",cc->base1);
   		rtl_printf("   PBOR: 0x%8.8x\n",cc->base2);
   		rtl_printf("   CSR : 0x%8.8x\n",reg->csr);
   		rtl_printf("   CNAF: 0x%8.8x\n",reg->cnaf);
   		rtl_printf("   FIFO: 0x%8.8x\n",reg->fifo);
		*/
 	}
  	else
    	rtl_printf("CAMAC: BaseAddress or IRQ Error\n");

  	if (rtl_register_chrdev(RTL_K2915_MAJOR, "ks2915", &cc_fops)) {
    	rtl_printf("CAMAC: Cannot register major device %d for k2915\n", 
			RTL_K2915_MAJOR);
    	return -EIO;
  		}
	
  	return 0;  
 }

void cleanup_module (void)
 {
  	struct cc_struct *cc = &cc_table[0];
  	if (MOD_IN_USE)
    	rtl_printf("CAMAC: K2915 busy (not removed).\n");
  	else {
    	rtl_unregister_chrdev(RTL_K2915_MAJOR, "ks2915");
    	release_region(cc->base1,IOR_SIZE);
    	release_region(cc->base2,PBOR_SIZE);
		rtl_free_global_irq(cc->irq);
    	rtl_printf("CAMAC: K2915 device driver removed.\n");
  		}
 }
#endif

static int cc_probe(struct cc_struct *cc_pci_hdr, struct camac *cc_reg)
 {
  	int value = 0;
  	struct pci_dev *pdev = NULL;
  	unsigned char pci_irq_line, pci_timer;

  	/* PCI Interface Operational Register (IOR) */
  	unsigned int pci_ior_addr;    
  	/* PCI Parallel Bus Operational Registers (PBOR) */
  	unsigned int pci_pbor_addr;
  
  	/*
   	* This driver will only work with one card, so quit
   	* after finding the first card and ignore the rest.
   	*/
  	if(pci_present()) {
		while( (pdev = pci_find_device(PCI_VENDOR_ID_KINETIC,
			PCI_DEVICE_ID_KINETIC_2915, pdev)) != NULL) {

      		/* One CANNOT set the IRQ here. It's set by the bios. */
      		pci_read_config_byte(pdev, PCI_INTERRUPT_LINE, 
		       &pci_irq_line);	        
      		pci_read_config_byte(pdev, PCI_LATENCY_TIMER, 
		       &pci_timer);
      		if (pci_timer != LTR_SET)
        		pci_write_config_byte(pdev, PCI_LATENCY_TIMER,LTR_SET);
                                
      		pci_ior_addr  = 
			   pdev->base_address[0] & PCI_BASE_ADDRESS_IO_MASK;
      		pci_pbor_addr = 
			   pdev->base_address[1] & PCI_BASE_ADDRESS_IO_MASK;
      
      		cc_pci_hdr->irq   = pci_irq_line;
      		cc_pci_hdr->base1 = pci_ior_addr;
      		cc_pci_hdr->base2 = pci_pbor_addr;

     		cc_reg->csr  += pci_pbor_addr;
      		cc_reg->cnaf += pci_pbor_addr;
      		cc_reg->tcr  += pci_pbor_addr;
      		cc_reg->srr  += pci_pbor_addr;

      		cc_reg->fifo += pci_ior_addr;
      		cc_reg->mwa  += pci_ior_addr;
      		cc_reg->mwtc += pci_ior_addr;
      		cc_reg->mra  += pci_ior_addr;
      		cc_reg->mrtc += pci_ior_addr;
      		cc_reg->ics  += pci_ior_addr;
      		cc_reg->bmcs += pci_ior_addr;
      
      		rtl_printf("CAMAC:K2915 found at ior = %#x,pbor = %#x,irq = %d.\n", 
					pci_ior_addr, pci_pbor_addr, pci_irq_line);
      
      		break;
    		}		
  		}
  	else
    	value = -1;
  
  	return value;
 }
